from typing import Dict, List, Literal
from pydantic import PrivateAttr

from .....utils.class_registry import register_class

from .....utils.desc_registry import DescDictType

from ....consts import DamageElementalType, DamageType

from ....modifiable_values import DamageValue

from ....action import Actions, MakeDamageAction

from ....struct import Cost

from ....card.equipment.artifact.ocean_hued import OceanHuedClam_4_2

from ....match import Match


class OceanHuedClam_4_3(OceanHuedClam_4_2):
    name: Literal["Ocean-Hued Clam"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(any_dice_number=3)
    _heal: int = PrivateAttr(2)

    def equip(self, match: Match) -> List[Actions]:
        super().equip(match)
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        # heal self
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        target_position=character.position,
                        damage_type=DamageType.HEAL,
                        damage=-self._heal,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            )
        ]


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Ocean-Hued Clam": {
        "descs": {
            "4.3": {
                "zh-CN": "入场时：治疗所附属角色2点。我方角色每受到3点治疗，此牌就累积1个「海染泡沫」。（最多累积2个）角色造成伤害时：消耗所有「海染泡沫」，每消耗1个都使造成的伤害+1。（角色最多装备1件「圣遗物」）",  # noqa: E501
                "en-US": "When played: Heal this character by 2 HP. For every 3 HP of healing your characters receive, this card accumulates 1 Sea-Dyed Foam (maximum of 2). When this character deals DMG: Consume all Sea-Dyed Foam. DMG is increased by 1 for each Sea-Dyed Foam consumed. (Characters can equip at most 1 Artifact)",  # noqa: E501
            }
        },
    },
}


register_class(OceanHuedClam_4_3, desc)



import logging
from typing import Literal
from lpsim.server.match import Match
from lpsim.utils import BaseModel
from lpsim.server.deck import Deck
from lpsim.agents import RandomAgent


class Main(BaseModel):
    """ """

    version: str = "1.0.0"
    name: Literal["GITCG"] = "GITCG"
    match: Match = Match()


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARNING)
    agent_0 = RandomAgent(player_idx=0)
    agent_1 = RandomAgent(player_idx=1)
    main = Main()
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Rhodeia of Loch
        character:Kamisato Ayaka
        character:Yaoyao
        Traveler's Handy Sword*5
        Gambler's Earrings*5
        Kanten Senmyou Blessing*5
        Sweet Madame*5
        Abyssal Summons*5
        Fatui Conspiracy*5
        Timmie*5
        """
    )
    main.match.set_deck([deck, deck])
    main.match.config.max_same_card_number = 30
    main.match.config.history_level = 10
    main.match.config.check_deck_restriction = False
    main.match.config.initial_hand_size = 20
    main.match.config.max_hand_size = 30
    main.match.config.card_number = None
    assert main.match.start()[0]
    main.match.step()

    while main.match.round_number < 100 and not main.match.is_game_end():
        if main.match.need_respond(0):
            current_agent = agent_0
        elif main.match.need_respond(1):
            current_agent = agent_1
        else:
            raise RuntimeError("no agent need to respond")
        resp = current_agent.generate_response(main.match)
        assert resp is not None
        main.match.respond(resp)
        main.match.step()

    main.match.get_history_json(filename="logs.txt")
    print("game end, save to logs.txt")



from typing import Any, Dict, List, Literal, Set, Type
from pydantic import PrivateAttr

from ....object_base import CreateSystemEventHandlerObject
from ....event_handler import SystemEventHandlerBase
from ....consts import ObjectPositionType
from ....modifiable_values import CostValue
from ....match import Match
from ....card.support.companions import CompanionBase
from ....card.support.locations import LocationBase
from ....card.equipment.artifact.base import ArtifactBase
from ....card.equipment.weapon.base import WeaponBase
from ....action import Actions
from ....event import UseCardEventArguments
from ....struct import Cost
from .....utils.class_registry import get_class_list_by_base_class, register_class
from .....utils.desc_registry import DescDictType
from ....card.support.items import RoundEffectItemBase


class MementoLensEventHandler_4_3(SystemEventHandlerBase):
    """
    Record used cards that matches type for both players.
    """

    name: Literal["Memento Lens"]
    version: Literal["4.3"] = "4.3"
    _accept_card_types: Type = PrivateAttr(
        WeaponBase | ArtifactBase | LocationBase | CompanionBase
    )
    _accept_class_name_set: Set[str] | None = PrivateAttr(None)
    card_used: List[Dict[str, int]] = [{}, {}]

    def _get_candidate_set(self, match: Match) -> Set[str]:
        if self._accept_class_name_set is not None:
            return self._accept_class_name_set
        kwargs: Dict[str, Any] = {}
        default_version = match.player_tables[
            self.position.player_idx
        ].player_deck_information.default_version
        if default_version is not None:
            kwargs["version"] = default_version
        candidate_list = get_class_list_by_base_class(self._accept_card_types, **kwargs)
        self._accept_class_name_set = set(candidate_list)
        return self._accept_class_name_set

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[Actions]:
        """
        If self played an accepted card type, record it.
        """
        player_idx = event.action.card_position.player_idx
        card_name = event.card_name
        if card_name not in self._get_candidate_set(match):
            # not accepted card type, do nothing
            return []
        self.card_used[player_idx][card_name] = 1
        return []


class MementoLens_4_3(CreateSystemEventHandlerObject, RoundEffectItemBase):
    name: Literal["Memento Lens"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1)
    max_usage_per_round: int = 1
    handler_name: str = "Memento Lens"

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        If self played a card that name in played_card_names, and has usage,
        decrease 2 cost.
        """
        if self.usage <= 0:
            # no usage, do nothing
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            source_area=ObjectPositionType.SUPPORT,
        ):
            # not self player or not in support, do nothing
            return value
        card = match.get_object(value.position)
        assert card is not None
        target_handler = self._get_event_handler(match)
        played_card_names = target_handler.card_used[self.position.player_idx]
        if card.name not in played_card_names:
            # not played card, do nothing
            return value
        # decrease cost
        result = [
            value.cost.decrease_cost(value.cost.elemental_dice_color),
            value.cost.decrease_cost(value.cost.elemental_dice_color),
        ]
        if any(result) and mode == "REAL":
            # decrease success and in real mode
            self.usage -= 1
        return value


desc: Dict[str, DescDictType] = {
    "SUPPORT/Memento Lens": {
        "names": {"en-US": "Memento Lens", "zh-CN": "留念镜"},
        "descs": {
            "4.3": {
                "en-US": "When you play a Weapon, Artifact, Location, or Companion card: If you have played a card with the same name previously during this match, spend 2 less Elemental Dice to play it. (Once per Round)\nUsage(s): 2",  # noqa: E501
                "zh-CN": "我方打出「武器」「圣遗物」「场地」「伙伴」手牌时：如果本场对局中我方曾经打出过所打出牌的同名卡牌，则少花费2个元素骰。（每回合1次）\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Prop_Mirror.png",  # noqa: E501
        "id": 323006,
    },
    "SYSTEM/Memento Lens": {
        "names": {"en-US": "Memento Lens", "zh-CN": "留念镜"},
        "descs": {"4.3": {}},
    },
}


register_class(MementoLens_4_3 | MementoLensEventHandler_4_3, desc)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..character_base import PhysicalNormalAttackBase
from ...struct import Cost
from ...consts import DieColor
from .beidou_3_8 import Beidou_3_8 as Beidou_3_8, LightningStorm_4_2
from .beidou_3_8 import Wavestrider as Wavestrider_3_8
from .beidou_3_8 import Stormbreaker as Stormbreaker_3_8
from .beidou_3_8 import Tidecaller


class Wavestrider(Wavestrider_3_8):
    damage: int = 2


class Stormbreaker(Stormbreaker_3_8):
    damage: int = 3
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=3
    )


class LightningStorm_3_4(LightningStorm_4_2):
    version: Literal["3.4"] = "3.4"
    need_to_activate: bool = True


class Beidou_3_4(Beidou_3_8):
    version: Literal["3.4"]
    skills: List[
        PhysicalNormalAttackBase | Tidecaller | Wavestrider | Stormbreaker
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Oceanborne",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Tidecaller(),
            Wavestrider(),
            Stormbreaker(),
        ]


register_class(Beidou_3_4 | LightningStorm_3_4)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import check_hp, make_respond, get_random_state


def test_fischl():
    """
    first: test equip talent, Oz will attack when normal attack, and this
    attack can go to next character, other character cannot trigger Oz attack,
    when run out of Oz, Oz will disappear immediately, Oz can trigger
    catalyzing field.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 1",
            "reroll pyro cryo hydro anemo",
            "sw_char 0 7",
            "card 1 0 2 3 4",
            "skill 0 1 2 3",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "skill 0 0 5 6",
            "end",
            "reroll 0 1 2 3 4 5 6 7",
            "tune 0 dendro",
            "tune 0 dendro",
            "tune 0 dendro",
            "skill 1 0 1 2",
            "tune 0 anemo",
            "skill 0 0 1 2",
            "end",
            "reroll 0 1 2 3 4 5 6 7",
            "tune 0 pyro",
            "tune 0 6",
            "skill 2 0 1 2",
            "sw_char 1 cryo",
            "sw_char 0 pyro",
            "end",
            "reroll 3 4 5 6 7",
            "skill 1 1 2 3",
            "sw_char 1 4",
            "skill 0 0 2 3",
            "sw_char 0 0",
            "end",
            "reroll",
            "skill 0 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "Fischl",
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "PyroMobMage",
                "element": "PYRO",
            },
        ],
        "cards": [
            {
                "name": "Stellar Predator",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 16:
                # Oz attack on next character
                check_hp(match, [[0, 8, 10], [10, 10, 10]])
                assert (
                    len(match.player_tables[0].characters[1].element_application) == 1
                )
                assert (
                    len(match.player_tables[0].characters[2].element_application) == 0
                )
                assert (
                    match.player_tables[0].characters[1].element_application[0]
                    == "ELECTRO"
                )
                # should have Oz and usage 2
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
                assert match.player_tables[1].summons[0].usage == 1
            elif len(agent_1.commands) == 11:
                # enemy hp 0 3 8 and only c1 have electro application
                check_hp(match, [[0, 3, 8], [10, 10, 10]])
                assert (
                    len(match.player_tables[0].characters[1].element_application) == 1
                )
                assert (
                    len(match.player_tables[0].characters[2].element_application) == 0
                )
                assert (
                    match.player_tables[0].characters[1].element_application[0]
                    == "ELECTRO"
                )
            elif len(agent_1.commands) == 4:
                # should have Oz and usage 2
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
                assert match.player_tables[1].summons[0].usage == 2
            elif len(agent_1.commands) == 1:
                # should have Oz
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 6
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    assert len(match.player_tables[1].dice.colors) == 5
    assert len(match.player_tables[1].summons) == 0

    assert match.state != MatchState.ERROR


def test_fischl_2():
    """
    second: when no talent or talent on other Fischl, cannot trigger Oz attack.
    re-summon Oz will refresh usage. Can equip talent multiple times.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll 4 5 6 7",
            "card 0 0 2 3 4",
            "sw_char 1 4",
            "tune 0 3",
            "skill 1 0 1 2",
            "end",
            "reroll 1 2 3 4 5 6 7",
            "skill 0 0 2 3",
            "sw_char 0 4",
            "tune 0 1",
            "tune 0 2",
            "card 0 0 0 1 2",
            "end",
            "reroll 0 1 2 3 4 5 6 7",
            "tune 0 7",
            "tune 0 7",
            "tune 0 7",
            "tune 0 7",
            "sw_char 1 7",
            "skill 0 0 5 6",
            "skill 2 0 1 2",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "Fischl",
            },
        ]
        * 3,
        "cards": [
            {
                "name": "Stellar Predator",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 16:
                # should be Oz and usage 2
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
                assert match.player_tables[1].summons[0].usage == 2
            elif len(agent_1.commands) == 10:
                # should be Oz and usage 1
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
                assert match.player_tables[1].summons[0].usage == 1
            elif len(agent_1.commands) == 9:
                # should be Oz and usage 1
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].name == "Oz"
                assert match.player_tables[1].summons[0].usage == 2
                check_hp(match, [[4, 10, 10], [10, 10, 10]])
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 8, 8], [10, 10, 10]])

    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_fischl()



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ELEMENT_TO_DIE_COLOR, ObjectPositionType

from ....action import CreateDiceAction, CreateObjectAction

from ....event import ReceiveDamageEventArguments, RoundPrepareEventArguments

from ....struct import Cost
from .base import ArtifactBase, RoundEffectArtifactBase


class GeneralsAncientHelm_3_5(ArtifactBase):
    name: Literal["General's Ancient Helm"]
    version: Literal["3.5"] = "3.5"
    cost: Cost = Cost(same_dice_number=2)
    usage: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create Unmovable Mountain for this character.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        position = self.position.set_area(ObjectPositionType.CHARACTER_STATUS)
        return [
            CreateObjectAction(
                object_name="Unmovable Mountain",
                object_position=position,
                object_arguments={},
            )
        ]


class TenacityOfTheMillelith_3_7(RoundEffectArtifactBase):
    name: Literal["Tenacity of the Millelith"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=3)
    max_usage_per_round: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        super().event_handler_ROUND_PREPARE(event, match)
        position = self.position.set_area(ObjectPositionType.CHARACTER_STATUS)
        return [
            CreateObjectAction(
                object_name="Unmovable Mountain",
                object_position=position,
                object_arguments={},
            )
        ]

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When this character received damage and is active character, create
        elemental die.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        if self.usage == 0:
            # no usage
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp == 0:
            # character is dying
            return []
        damage = event.final_damage
        if not self.position.check_position_valid(
            damage.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_is_active_character=True,
        ):
            # damage not attack self, or self not active character
            return []
        # create die
        self.usage -= 1
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=ELEMENT_TO_DIE_COLOR[character.element],
            )
        ]


register_class(GeneralsAncientHelm_3_5 | TenacityOfTheMillelith_3_7)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..character_base import PhysicalNormalAttackBase

from ...consts import DieColor

from ...struct import Cost
from .jean_3_3 import (
    Jean_3_3,
    DandelionField_3_3,
    LandsOfDandelion_3_3,
    GaleBlade,
    DandelionBreeze as DB_3_3,
)


class DandelionField_4_2(DandelionField_3_3):
    version: Literal["4.2"] = "4.2"
    damage: int = 1


class LandsOfDandelion_4_2(LandsOfDandelion_3_3):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=4, charge=2
    )


class DandelionBreeze(DB_3_3):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=4, charge=2
    )


class Jean_4_2(Jean_3_3):
    version: Literal["4.2"] = "4.2"
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | GaleBlade | DandelionBreeze] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Favonius Bladework",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GaleBlade(),
            DandelionBreeze(),
        ]


register_class(Jean_4_2 | DandelionField_4_2 | LandsOfDandelion_4_2)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import DamageValue
from ....action import ActionTypes, CreateObjectAction, MakeDamageAction
from ....consts import DamageElementalType, DamageType, IconType
from ....status.character_status.base import RoundEndAttackCharacterStatus
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....card.event.foods import FoodCardBase
from .....utils.desc_registry import DescDictType


class MatsutakeMeatRollsStatus_4_4(RoundEndAttackCharacterStatus):
    name: Literal["Matsutake Meat Rolls"] = "Matsutake Meat Rolls"
    version: Literal["4.4"] = "4.4"
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.HEAL] = IconType.HEAL
    damage: int = -1
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL


class MatsutakeMeatRolls_4_4(FoodCardBase):
    name: Literal["Matsutake Meat Rolls"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost(same_dice_number=2)

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=target,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        )
        create_1 = ret[0]
        assert create_1.type == ActionTypes.CREATE_OBJECT
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=create_1.object_position,
                object_arguments={},
            )
        )
        return ret


desc: Dict[str, DescDictType] = {
    "CHARACTER_STATUS/Matsutake Meat Rolls": {
        "names": {"en-US": "Matsutake Meat Rolls", "zh-CN": "松茸酿肉卷"},
        "descs": {
            "4.4": {
                "en-US": "For the next 3 Rounds, heal this character for 1 HP again during the End Phase. (Usages: 3)",  # noqa: E501
                "zh-CN": "结束阶段治疗此角色1点。（可用次数：3）",  # noqa: E501
            }
        },
    },
    "CARD/Matsutake Meat Rolls": {
        "names": {"en-US": "Matsutake Meat Rolls", "zh-CN": "松茸酿肉卷"},
        "descs": {
            "4.4": {
                "en-US": "Heal the target character for 2 HP. For the next 3 Rounds, heal this character for 1 HP again during the End Phase.\n(A character can consume at most 1 Food per Round)",  # noqa: E501
                "zh-CN": "治疗目标角色2点，3回合内的结束阶段再治疗此角色1点。\n（每回合每个角色最多食用1次「料理」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Food_Matsutake.png",  # noqa: E501
        "id": 333014,
    },
}


register_class(MatsutakeMeatRolls_4_4 | MatsutakeMeatRollsStatus_4_4, desc)



from .card import (
    LocationBase,
    ItemBase,
    WeaponBase,
    SupportBase,
    ArtifactBase,
    CompanionBase,
)
from .character import CharacterBase, TalentBase, SkillBase
from .status import StatusBase, CharacterStatusBase, TeamStatusBase
from .summon import SummonBase
from .match import Match, MatchState
from .object_base import ObjectBase, CardBase
from .deck import Deck

# import patch to automatically register classes
from .patch import *  # noqa: F401, F403

__all__ = (
    "LocationBase",
    "ItemBase",
    "WeaponBase",
    "SupportBase",
    "ArtifactBase",
    "CompanionBase",
    "CharacterBase",
    "TalentBase",
    "SkillBase",
    "StatusBase",
    "CharacterStatusBase",
    "TeamStatusBase",
    "SummonBase",
    "Match",
    "MatchState",
    "ObjectBase",
    "CardBase",
    "Deck",
)



from .server.deck import Deck  # noqa: F401
from .server.match import Match, MatchState  # noqa: F401
from .utils import (  # noqa: F401
    register_class,
    DescDictType,
    update_desc,
    deck_str_to_deck_code,
    deck_code_to_deck_str,
)


try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ModuleNotFoundError:
    __version__ = "unknown"
    __version_tuple__ = (0, 0, 0)



from ...utils import import_all_modules
from .character_base import CharacterBase, TalentBase, SkillBase


import_all_modules(__file__, __name__, exceptions={"template"})
__all__ = ("CharacterBase", "TalentBase", "SkillBase")



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear


# Skills


class SacredRiteHeronsSanctum(ElementalSkillBase):
    name: Literal["Sacred Rite: Heron's Sanctum"] = "Sacred Rite: Heron's Sanctum"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        prepare
        """
        return [
            self.charge_self(1),
            self.create_character_status("Heron Shield"),
        ]


class HeronStrike(ElementalSkillBase):
    name: Literal["Heron Strike"] = "Heron Strike"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        always invalid to directly use.
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        no charge, only attack
        """
        return [self.attack_opposite_active(match, self.damage, self.damage_type)]


class SacredRiteWagtailsTide(ElementalBurstBase):
    name: Literal["Sacred Rite: Wagtail's Tide"] = "Sacred Rite: Wagtail's Tide"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_team_status("Prayer of the Crimson Crown"),
            ],
        )


# Talents


class TheOverflow_3_8(SkillTalent):
    name: Literal["The Overflow"]
    version: Literal["3.8"]
    character_name: Literal["Candace"] = "Candace"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Sacred Rite: Wagtail's Tide"] = "Sacred Rite: Wagtail's Tide"


class TheOverflow_4_2(TheOverflow_3_8):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )


# character base


class Candace_3_8(CharacterBase):
    name: Literal["Candace"]  # Do not set default value for character name
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | SacredRiteHeronsSanctum
        | SacredRiteWagtailsTide
        | HeronStrike
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Gleaming Spear - Guardian Stance",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SacredRiteHeronsSanctum(),
            SacredRiteWagtailsTide(),
            HeronStrike(),
        ]


register_class(Candace_3_8 | TheOverflow_3_8 | TheOverflow_4_2)



from typing import Dict, Literal

from .....utils.class_registry import register_class

from .....utils.desc_registry import DescDictType
from ....struct import Cost
from ....card.event.arcane_legend import JoyousCelebration_4_2


class JoyousCelebration_4_3(JoyousCelebration_4_2):
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(arcane_legend=True)


desc: Dict[str, DescDictType] = {
    "ARCANE/Joyous Celebration": {
        "descs": {
            "4.3": {
                "zh-CN": "$ARCANE/Joyous Celebration|descs|4.2|zh-CN",
                "en-US": "$ARCANE/Joyous Celebration|descs|4.2|en-US",
            }
        },
    },
}


register_class(JoyousCelebration_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost

from ...consts import DamageElementalType, DamageType, IconType, SkillType

from ...modifiable_values import DamageDecreaseValue, DamageValue

from ...action import MakeDamageAction

from ...event import PlayerActionStartEventArguments
from .base import ExtraAttackTeamStatus, RoundTeamStatus, UsageTeamStatus


class TenkoThunderbolts_3_7(UsageTeamStatus):
    name: Literal["Tenko Thunderbolts"] = "Tenko Thunderbolts"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_PLAYER_ACTION_START(
        self, event: PlayerActionStartEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        Attack opponent active character
        """
        if self.position.player_idx != event.player_idx:
            # not our turn
            return []
        target_table = match.player_tables[1 - self.position.player_idx]
        target_character_idx = target_table.active_character_idx
        target_character = target_table.characters[target_character_idx]
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target_character.position,
                        damage=3,
                        damage_elemental_type=DamageElementalType.ELECTRO,
                        cost=Cost(),
                    )
                ],
            )
        ]


class ThunderbeastsTarge_3_4(RoundTeamStatus, ExtraAttackTeamStatus):
    name: Literal["Thunderbeast's Targe"] = "Thunderbeast's Targe"
    version: Literal["3.4"] = "3.4"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    trigger_skill_type: SkillType | None = SkillType.NORMAL_ATTACK
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    decrease_usage: bool = False

    def value_modifier_DAMAGE_DECREASE(
        self,
        value: DamageDecreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageDecreaseValue:
        """
        If this character receives damage, and not piercing, and greater than
        3, decrease damage by 1
        """
        if (
            value.is_corresponding_character_receive_damage(
                self.position,
                match,
            )
            and value.damage >= 3
        ):
            value.damage -= 1
        return value


register_class(TenkoThunderbolts_3_7 | ThunderbeastsTarge_3_4)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageIncreaseValue

from ....action import CreateObjectAction

from ....event import SkillEndEventArguments

from .base import RoundEffectWeaponBase, WeaponBase

from ....struct import Cost, ObjectPosition

from ....consts import ObjectPositionType, ObjectType, WeaponType


class WolfsGravestone_3_3(WeaponBase):
    name: Literal["Wolf's Gravestone"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["3.3"] = "3.3"
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert value.target_position.area == ObjectPositionType.CHARACTER
        character = match.get_object(value.target_position)
        if character.hp <= 6:
            self.damage_increase = 3
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        self.damage_increase = 1
        return value


class TheBell_3_7(RoundEffectWeaponBase):
    name: Literal["The Bell"]
    type: Literal[ObjectType.WEAPON] = ObjectType.WEAPON
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.CLAYMORE

    cost: Cost = Cost(same_dice_number=3)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self character use any skill, and have usage, create Rebellious
        Shield.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self character use skill
            return []
        if self.usage == 0:
            # no usage
            return []
        self.usage -= 1
        return [
            CreateObjectAction(
                object_name="Rebellious Shield",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=-1,
                ),
                object_arguments={},
            )
        ]


register_class(WolfsGravestone_3_3 | TheBell_3_7)



from typing import Dict, List, Literal

from pydantic import PrivateAttr

from ....action import (
    ActionTypes,
    Actions,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ....event import (
    CreateObjectEventArguments,
    RemoveObjectEventArguments,
    RoundPrepareEventArguments,
)
from .....utils.class_registry import register_class
from ....status.character_status.base import (
    PrepareCharacterStatus,
    UsageCharacterStatus,
)
from ....modifiable_values import DamageIncreaseValue, DamageValue
from ....match import Match
from ....struct import Cost
from ....consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    ObjectType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class TotalCollapse_4_3(UsageCharacterStatus):
    name: Literal["Total Collapse"] = "Total Collapse"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.DEBUFF

    available_handler_in_trashbin: List[ActionTypes] = [ActionTypes.REMOVE_OBJECT]

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Match,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if (
            value.damage_elemental_type != DamageElementalType.PHYSICAL
            and value.damage_elemental_type != DamageElementalType.ANEMO
        ):
            # not right damage type
            return value
        if value.is_corresponding_character_receive_damage(self.position, match):
            # this character receive damage
            assert mode == "REAL"
            self.usage -= 1
            value.damage += 2
        return value


class PerpetualCleansingStatus_4_3(PrepareCharacterStatus):
    name: Literal["Perpetual Cleansing"] = "Perpetual Cleansing"
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Dvalin"] = "Dvalin"
    skill_name: Literal["Perpetual Cleansing"] = "Perpetual Cleansing"


class UltimateCleansingStatus_4_3(PrepareCharacterStatus):
    name: Literal["Ultimate Cleansing"] = "Ultimate Cleansing"
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Dvalin"] = "Dvalin"
    skill_name: Literal["Ultimate Cleansing"] = "Ultimate Cleansing"


class TempestuousBarrage(ElementalSkillBase):
    name: Literal["Tempestuous Barrage"] = "Tempestuous Barrage"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(
            match, [self.create_opposite_character_status(match, "Total Collapse", {})]
        )


class DvalinsCleansing(ElementalSkillBase):
    name: Literal["Dvalin's Cleansing"] = "Dvalin's Cleansing"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=5,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(
            match, [self.create_character_status("Perpetual Cleansing", {})]
        )


class PerpetualCleansing(ElementalSkillBase):
    name: Literal["Perpetual Cleansing"] = "Perpetual Cleansing"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost()

    def is_valid(self, match: Match) -> bool:
        return False

    def get_actions(self, match: Match) -> List[Actions]:
        table = match.player_tables[1 - self.position.player_idx]
        target = table.next_character_idx()
        if target is None:
            target = table.active_character_idx
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=table.characters[target].position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_type,
                        cost=self.cost.copy(),
                    )
                ],
                create_objects=[self.create_character_status("Ultimate Cleansing", {})],
            ),
        ]


class UltimateCleansing(ElementalSkillBase):
    name: Literal["Ultimate Cleansing"] = "Ultimate Cleansing"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost()

    def is_valid(self, match: Match) -> bool:
        return False

    def get_actions(self, match: Match) -> List[Actions]:
        table = match.player_tables[1 - self.position.player_idx]
        target = table.previous_character_idx()
        if target is None:
            target = table.active_character_idx
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=table.characters[target].position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_type,
                        cost=self.cost.copy(),
                    )
                ]
            ),
        ]


class CaelestinumFinaleTermini(ElementalBurstBase):
    name: Literal["Caelestinum Finale Termini"] = "Caelestinum Finale Termini"
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        table = match.player_tables[1 - self.position.player_idx]
        for character in table.characters:
            if table.active_character_idx == character.position.character_idx:
                continue
            if character.is_defeated:
                continue
            # apply status
            damage_action.create_objects.append(
                CreateObjectAction(
                    object_name="Total Collapse",
                    object_position=character.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_arguments={},
                )
            )
        return ret


class RendingVortex_4_3(SkillTalent):
    name: Literal["Rending Vortex"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Dvalin"] = "Dvalin"
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)
    skill: Literal["Tempestuous Barrage"] = "Tempestuous Barrage"
    round_usage: int = 1
    _max_round_usage: int = PrivateAttr(1)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.round_usage = self._max_round_usage
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        when Total collapse is renew create, treat it as old one removed.
        """
        if event.create_result != "RENEW":
            # not renew, do nothing
            return []
        # make fake event arguments
        fake_event = RemoveObjectEventArguments(
            action=RemoveObjectAction(
                object_position=event.action.object_position,
            ),
            object_name=event.action.object_name,
            # no effect, set one default
            object_type=ObjectType.CHARACTER_STATUS,
        )
        ret = self.event_handler_REMOVE_OBJECT(fake_event, match)
        return ret

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        When opposite Total Collapse removed, and opposite is active, and found
        self is in character, create Total Collapse on next standby character.
        """
        if self.round_usage <= 0:
            # no usage
            return []
        if event.object_name != "Total Collapse":
            # not Total Collapse
            return []
        if not self.position.check_position_valid(
            event.action.object_position,
            match,
            player_idx_same=False,
            source_area=ObjectPositionType.CHARACTER,
            target_is_active_character=True,
        ):
            # not from opposite active character, or self not equipped
            return []
        # is opposite active, append Total Collapse to next standby character
        table = match.player_tables[1 - self.position.player_idx]
        target = table.next_character_idx()
        if target is None:
            # no target, do nothing
            return []
        self.round_usage -= 1
        target_character = table.characters[target]
        return [
            CreateObjectAction(
                object_name="Total Collapse",
                object_position=target_character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]


class Dvalin_4_3(CharacterBase):
    name: Literal["Dvalin"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | TempestuousBarrage
        | DvalinsCleansing
        | PerpetualCleansing
        | UltimateCleansing
        | CaelestinumFinaleTermini
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Lacerating Slash",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TempestuousBarrage(),
            DvalinsCleansing(),
            PerpetualCleansing(),
            UltimateCleansing(),
            CaelestinumFinaleTermini(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Dvalin": {
        "names": {"en-US": "Dvalin", "zh-CN": "特瓦林"},
        "image_path": "cardface/Char_Monster_Dvalin.png",  # noqa: E501
        "id": 2502,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Dvalin_NORMAL_ATTACK/Lacerating Slash": {
        "names": {"en-US": "Lacerating Slash", "zh-CN": "裂爪横击"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Dvalin_ELEMENTAL_SKILL/Tempestuous Barrage": {
        "names": {"en-US": "Tempestuous Barrage", "zh-CN": "暴风轰击"},
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Anemo DMG. The target character receives Total Collapse.",  # noqa: E501
                "zh-CN": "造成2点风元素伤害，目标角色附属坍毁。",
            }
        },
    },
    "CHARACTER_STATUS/Total Collapse": {
        "names": {"en-US": "Total Collapse", "zh-CN": "坍毁"},
        "descs": {
            "4.3": {
                "en-US": "Physical DMG or Anemo DMG taken by the affected character increased by 2.\nUsage(s): 1",  # noqa: E501
                "zh-CN": "所附属角色受到的物理伤害或风元素伤害+2。\n可用次数：1",
            }
        },
    },
    "SKILL_Dvalin_ELEMENTAL_SKILL/Dvalin's Cleansing": {
        "names": {"en-US": "Dvalin's Cleansing", "zh-CN": "风龙涤流"},
        "descs": {
            "4.3": {
                "en-US": 'Deals 2 Anemo DMG and then separately performs "Prepare Skill|s1:prepares" for Perpetual Cleansing and Ultimate Cleansing.',  # noqa: E501
                "zh-CN": "造成2点风元素伤害，然后分别准备技能：长延涤流和终幕涤流。",
            }
        },
    },
    "CHARACTER_STATUS/Perpetual Cleansing": {
        "names": {"en-US": "Perpetual Cleansing", "zh-CN": "长延涤流"},
        "descs": {
            "4.3": {
                "en-US": "Elemental Skill\n(Prepare for 1 turn)\nDeals 1 Anemo DMG the next opposing character on standby. After this, Prepare Skill|s1:prepares: Ultimate Cleansing. (If there are no opposing characters on standby, deals DMG to active character instead)",  # noqa: E501
                "zh-CN": "元素战技\n（需准备1个行动轮）\n对下一个敌方后台角色造成1点风元素伤害，然后准备技能：终幕涤流。（敌方没有后台角色时，改为对出战角色造成伤害）",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Ultimate Cleansing": {
        "names": {"en-US": "Ultimate Cleansing", "zh-CN": "终幕涤流"},
        "descs": {
            "4.3": {
                "en-US": "Elemental Skill\n(Prepare for 1 turn)\nDeals 2 Anemo DMG the previous opposing character on standby. (If there are no opposing characters on standby, deals DMG to the active character instead)",  # noqa: E501
                "zh-CN": "元素战技\n（需准备1个行动轮）\n对上一个敌方后台角色造成2点风元素伤害。（敌方没有后台角色时，改为对出战角色造成伤害）",  # noqa: E501
            }
        },
    },
    "SKILL_Dvalin_ELEMENTAL_BURST/Caelestinum Finale Termini": {
        "names": {"en-US": "Caelestinum Finale Termini", "zh-CN": "终天闭幕曲"},
        "descs": {
            "4.3": {
                "en-US": "Deals 5 Anemo DMG. Applies Total Collapse to all opposing standby characters.",  # noqa: E501
                "zh-CN": "造成5点风元素伤害，所有敌方后台角色附属坍毁。",
            }
        },
    },
    "SKILL_Dvalin_ELEMENTAL_SKILL/Perpetual Cleansing": {
        "names": {"en-US": "Perpetual Cleansing", "zh-CN": "长延涤流"},
        "descs": {
            "4.3": {
                "en-US": "(Prepare for 1 turn)\nDeals 1 Anemo DMG the next opposing character on standby. After this, Prepare Skill|s1:prepares: Ultimate Cleansing. (If there are no opposing characters on standby, deals DMG to active character instead)",  # noqa: E501
                "zh-CN": "（需准备1个行动轮）\n对下一个敌方后台角色造成1点风元素伤害，然后准备技能：终幕涤流。（敌方没有后台角色时，改为对出战角色造成伤害）",  # noqa: E501
            }
        },
    },
    "SKILL_Dvalin_ELEMENTAL_SKILL/Ultimate Cleansing": {
        "names": {"en-US": "Ultimate Cleansing", "zh-CN": "终幕涤流"},
        "descs": {
            "4.3": {
                "en-US": "(Prepare for 1 turn)\nDeals 2 Anemo DMG the previous opposing character on standby. (If there are no opposing characters on standby, deals DMG to the active character instead)",  # noqa: E501
                "zh-CN": "（需准备1个行动轮）\n对上一个敌方后台角色造成2点风元素伤害。（敌方没有后台角色时，改为对出战角色造成伤害）",  # noqa: E501
            }
        },
    },
    "TALENT_Dvalin/Rending Vortex": {
        "names": {"en-US": "Rending Vortex", "zh-CN": "毁裂风涡"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Dvalin, equip this card.\nAfter Dvalin equips this card, immediately use Tempestuous Barrage once.\nWhen your Dvalin, who has this card equipped, is on the field, when Total Collapse attached to opposing active character is removed: Apply Total Collapse to the next opposing standby character. (Once per Round)\n(You must have Dvalinin your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为特瓦林时，装备此牌。\n特瓦林装备此牌后，立刻使用一次暴风轰击。\n装备有此牌的特瓦林在场时，敌方出战角色所附属的坍毁状态被移除后：对下一个敌方后台角色附属坍毁。（每回合1次）\n（牌组中包含特瓦林，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Dvalin.png",  # noqa: E501
        "id": 225021,
    },
}


register_class(
    TotalCollapse_4_3
    | PerpetualCleansingStatus_4_3
    | UltimateCleansingStatus_4_3
    | RendingVortex_4_3
    | Dvalin_4_3,
    desc,
)



from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundCharacterStatus, ShieldCharacterStatus


class UnmovableMountain_3_5(ShieldCharacterStatus):
    name: Literal["Unmovable Mountain"] = "Unmovable Mountain"
    version: Literal["3.5"] = "3.5"
    usage: int = 2
    max_usage: int = 2


class VermillionHereafter_4_0(RoundCharacterStatus):
    name: Literal["Vermillion Hereafter"]
    version: Literal["4.0"] = "4.0"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not current character using normal attack
            return value
        # increase damage
        assert mode == "REAL"
        value.damage += 1
        return value


register_class(UnmovableMountain_3_5 | VermillionHereafter_4_0)



# type: ignore


from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue
from ...event import ReceiveDamageEventArguments, RoundPrepareEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    CostLabels,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.
# Round status, will last for several rounds and disappear
# Skills


class Tidecaller(ElementalSkillBase):
    name: Literal["Tidecaller"] = "Tidecaller"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create status
        """
        return [
            self.charge_self(1),
            self.create_character_status("Tidecaller: Surf Embrace"),
        ]


class Wavestrider(ElementalSkillBase):
    name: Literal["Wavestrider"] = "Wavestrider"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        always invalid to directly use.
        """
        return False

    def get_actions(self, match: Any) -> List[MakeDamageAction]:
        return [self.attack_opposite_active(match, self.damage, self.damage_type)]


class Stormbreaker(ElementalBurstBase):
    name: Literal["Stormbreaker"] = "Stormbreaker"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(
            match,
            [
                self.create_team_status("Thunderbeast's Targe"),
            ],
        )


# Talents


class LightningStorm_4_2(SkillTalent):
    name: Literal["Lightning Storm"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Beidou"] = "Beidou"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Tidecaller"] = "Tidecaller"

    usage: int = 2
    max_usage: int = 2
    activated: bool = False
    need_to_activate: bool = False

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset usage and activate
        """
        self.usage = self.max_usage
        self.activated = False
        return []

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If self receive damage and has status, activate
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        if not event.final_damage.is_corresponding_character_receive_damage(
            self.position,
            match,
        ):
            # not corresponding character
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        for status in character.status:
            if status.name == "Tidecaller: Surf Embrace":
                self.activated = True
                break
        return []

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        If activated, decrease self normal attack cost by 1
        """
        if not self.activated and self.need_to_activate:
            # not activated and need to activate
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self skill
            return value
        if value.cost.label & CostLabels.NORMAL_ATTACK.value == 0:
            # not normal attack
            return value
        if self.usage <= 0:
            # no usage
            return value
        # decrease
        if value.cost.decrease_cost(None):  # pragma: no branch
            if mode == "REAL":
                self.usage -= 1
        return value


# character base


class Beidou_3_8(CharacterBase):
    name: Literal["Beidou"]
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | Tidecaller | Wavestrider | Stormbreaker
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Oceanborne",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Tidecaller(),
            Wavestrider(),
            Stormbreaker(),
        ]


register_class(Beidou_3_8 | LightningStorm_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageValue
from ...event import RoundEndEventArguments, RoundPrepareEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


class MelodyLoop_3_3(AttackerSummonBase):
    name: Literal["Melody Loop"] = "Melody Loop"
    version: Literal["3.3"] = "3.3"
    damage_elemental_type: DamageElementalType = DamageElementalType.HEAL
    damage: int = -1
    usage: int = 2
    max_usage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        player_idx = self.position.player_idx
        assert self.usage > 0
        self.usage -= 1
        target_table = match.player_tables[player_idx]
        damage_action = MakeDamageAction(
            damage_value_list=[],
        )
        for cid, character in enumerate(target_table.characters):
            if character.is_alive:
                damage_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                )
            if target_table.active_character_idx == cid:
                assert character.is_alive
                damage_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.ELEMENT_APPLICATION,
                        target_position=character.position,
                        damage=0,
                        damage_elemental_type=DamageElementalType.HYDRO,
                        cost=Cost(),
                    )
                )
        return [damage_action]


class LetTheShowBegin(ElementalSkillBase):
    name: Literal["Let the Show Begin♪"] = "Let the Show Begin♪"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_number=3, elemental_dice_color=DieColor.HYDRO)

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match, [self.create_summon("Melody Loop")])
        return ret


class ShiningMiracle(ElementalBurstBase):
    name: Literal["Shining Miracle"] = "Shining Miracle"
    damage: int = -4
    damage_type: DamageElementalType = DamageElementalType.HEAL
    cost: Cost = Cost(
        elemental_dice_number=3, elemental_dice_color=DieColor.HYDRO, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = [self.charge_self(-3)]
        characters = match.player_tables[self.position.player_idx].characters
        heal_action = MakeDamageAction(
            damage_value_list=[],
        )
        for character in characters:
            if character.is_alive:
                heal_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_type,
                        cost=self.cost.copy(),
                    )
                )
        ret.append(heal_action)
        return ret


class GloriousSeason_3_3(SkillTalent):
    name: Literal["Glorious Season"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Barbara"] = "Barbara"
    cost: Cost = Cost(
        elemental_dice_number=4,
        elemental_dice_color=DieColor.HYDRO,
    )
    skill: Literal["Let the Show Begin♪"] = "Let the Show Begin♪"
    usage: int = 1

    def event_handler_ROUND_PREPARE(
        self, match: Any, event_args: RoundPrepareEventArguments
    ) -> List[Actions]:
        """reset usage"""
        self.usage = 1
        return []

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        Once per round, if summon "Melody Loop" is valid in our summon area,
        and our do switch, and have cost, then cost -1.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, no effect
            return value
        if self.usage <= 0:
            # no usage, no effect
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER == 0:
            # not switch charatctor, no effect
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self player, no effect
            return value
        summons = match.player_tables[self.position.player_idx].summons
        have_summon = False
        for summon in summons:
            if summon.name == "Melody Loop":
                have_summon = True
                break
        if not have_summon:
            # not have summon, no effect
            return value
        # decrease 1 any cost
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value


class GloriousSeason_4_2(GloriousSeason_3_3):
    version: Literal["4.2"] = "4.2"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
    )


class Barbara_3_3(CharacterBase):
    name: Literal["Barbara"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[ElementalNormalAttackBase | LetTheShowBegin | ShiningMiracle] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Whisper of Water",
                damage_type=DamageElementalType.HYDRO,
                cost=ElementalNormalAttackBase.get_cost(ElementType.HYDRO),
            ),
            LetTheShowBegin(),
            ShiningMiracle(),
        ]


register_class(Barbara_3_3 | GloriousSeason_3_3 | GloriousSeason_4_2 | MelodyLoop_3_3)



from typing import Any, Literal, List

from ....utils.class_registry import register_class

from ...event import CreateObjectEventArguments

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction

from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)

from ..character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    SkillTalent,
)


class AllSchemesToKnow(ElementalSkillBase):
    name: Literal["All Schemes to Know"] = "All Schemes to Know"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[CreateObjectAction] = []
        # get target characters and active character.
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_characters = target_table.characters
        target_active_character = target_characters[target_table.active_character_idx]
        # default set status to active character
        position = target_active_character.position.set_area(
            ObjectPositionType.CHARACTER_STATUS
        )
        ret.append(
            CreateObjectAction(
                object_name="Seed of Skandha",
                object_position=position,
                object_arguments={},
            )
        )
        # get seed of skandha status.
        active_character_status_names = [x.name for x in target_active_character.status]
        if "Seed of Skandha" in active_character_status_names:
            # apply to all opposing characters.
            for character in target_characters:
                if character.is_defeated:
                    continue
                if character.id != target_active_character.id:
                    position = character.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    )
                    ret.append(
                        CreateObjectAction(
                            object_name="Seed of Skandha",
                            object_position=position,
                            object_arguments={},
                        )
                    )
        # apply damage after apply status.
        return super().get_actions(match, ret)


class AllSchemesToKnowTathata(ElementalSkillBase):
    name: Literal["All Schemes to Know: Tathata"] = "All Schemes to Know: Tathata"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=5,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[CreateObjectAction] = []
        # get target characters and active character.
        player_idx = self.position.player_idx
        target_table = match.player_tables[1 - player_idx]
        target_characters = target_table.characters
        # apply to all opposing characters.
        for character in target_characters:
            if character.is_defeated:
                continue
            position = character.position.set_area(ObjectPositionType.CHARACTER_STATUS)
            ret.append(
                CreateObjectAction(
                    object_name="Seed of Skandha",
                    object_position=position,
                    object_arguments={},
                )
            )
        return super().get_actions(match, ret)


class IllusoryHeart(ElementalBurstBase):
    name: Literal["Illusory Heart"] = "Illusory Heart"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
        charge=2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match, [self.create_team_status("Shrine of Maya")])


class TheSeedOfStoredKnowledge_3_7(SkillTalent):
    """
    Three effects are implemented in different places:
        Pyro: when Seed of Skandha is triggered, it will check whether need
            to add dendro damage.
        Electro: when Shrine of Maya is summoned, the status will check whether
            need to add one usage to Seed of Skandha.
        Hydro: when Shrine of Maya is summoned, this talent card will try
            to add one usage if is equipped.
    """

    name: Literal["The Seed of Stored Knowledge"]
    character_name: Literal["Nahida"] = "Nahida"
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
        charge=2,
    )
    skill: Literal["Illusory Heart"] = "Illusory Heart"

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        when Shrine of Maya is summoned, if this card is equipped, add one
        usage.
        """
        if event.action.object_name != "Shrine of Maya":
            # not creating Shrine of Maya, do nothing.
            return []
        if not self.position.check_position_valid(
            event.action.object_position,
            match,
            player_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
            source_is_active_character=True,
        ):
            # not self, or not equipped, or not active character, do nothing.
            return []
        table = match.player_tables[self.position.player_idx]
        has_hydro_character = False
        for character in table.characters:
            if character.element == ElementType.HYDRO:
                has_hydro_character = True
                break
        if not has_hydro_character:
            # no hydro character, do nothing.
            return []
        team_status = table.team_status
        for status in team_status:
            if status.name == "Shrine of Maya":
                # has Shrine Of Maya, add one usage.
                return [
                    ChangeObjectUsageAction(
                        object_position=status.position,
                        change_usage=1,
                    )
                ]
        raise AssertionError("Shrine of Maya not found.")


class Nahida_3_7(CharacterBase):
    name: Literal["Nahida"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | AllSchemesToKnow
        | AllSchemesToKnowTathata
        | IllusoryHeart
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU,
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Akara",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            AllSchemesToKnow(),
            AllSchemesToKnowTathata(),
            IllusoryHeart(),
        ]


register_class(Nahida_3_7 | TheSeedOfStoredKnowledge_3_7)



from typing import Any, Dict, List, Literal, Type
from pydantic import PrivateAttr

from ....action import ActionTypes, Actions, CreateObjectAction, RemoveObjectAction
from .....utils.class_registry import get_instance, register_class
from ....event import GameStartEventArguments, MoveObjectEventArguments
from ....modifiable_values import CostValue
from ....status.team_status.base import RoundTeamStatus, UsageTeamStatus
from ....consts import CostLabels, IconType, ObjectPositionType
from .....utils.class_registry import get_class_list_by_base_class
from ....card.support.base import SupportBase
from ....struct import Cost, ObjectPosition
from ....match import Match
from ....object_base import EventCardBase
from .....utils.desc_registry import DescDictType


class SunyataFlowerStatus_4_4(UsageTeamStatus, RoundTeamStatus):
    name: Literal["Sunyata Flower"] = "Sunyata Flower"
    version: Literal["4.4"] = "4.4"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF
    target_label: int = CostLabels.SUPPORTS.value

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        decrease weapons or artifact cost by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not self character, do nothing
            return value
        if value.cost.label & self.target_label == 0:
            # not weapon or artifact, do nothing
            return value
        # try decrease once
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == "REAL":
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping support,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class SunyataFlower_4_4(EventCardBase):
    name: Literal["Sunyata Flower"]
    version: Literal["4.4"] = "4.4"
    cost: Cost = Cost()
    _accept_card_types: Type = PrivateAttr(SupportBase)
    _accept_class_list: List[str] | None = PrivateAttr(None)
    _accept_version: str | None = PrivateAttr(None)

    available_handler_in_deck: List[ActionTypes] = [ActionTypes.GAME_START]

    def _get_candidate_list(self, match: Match) -> List[str]:
        if self._accept_class_list is not None:
            return self._accept_class_list
        kwargs: Dict[str, Any] = {"exclude": set([self.name])}
        default_version = match.player_tables[
            self.position.player_idx
        ].player_deck_information.default_version
        if default_version is not None:
            kwargs["version"] = default_version
        self._accept_version = default_version
        candidate_list = get_class_list_by_base_class(self._accept_card_types, **kwargs)
        self._accept_class_list = candidate_list
        return candidate_list

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Match
    ) -> List[Actions]:
        """
        check all cards that able to generate. If they contains GAME_START
        event and available in deck, call them.
        """
        candidate_list = self._get_candidate_list(match)
        args = {}
        if self._accept_version is not None:
            args["version"] = self._accept_version
        candidate_instance = [
            get_instance(self._accept_card_types, {"name": name, **args})
            for name in candidate_list
        ]
        res: List[Actions] = []
        for i in candidate_instance:
            if ActionTypes.GAME_START in i.available_handler_in_deck:
                res += i.event_handler_GAME_START(event, match)
        return res

    def is_valid(self, match: Match) -> bool:
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        """
        Our supports
        """
        supports = match.player_tables[self.position.player_idx].supports
        return [x.position for x in supports]

    def get_actions(
        self, target: ObjectPosition | None, match: Match
    ) -> List[RemoveObjectAction | CreateObjectAction]:
        """
        remove target, and generate 2 random support cards, and create status
        """
        assert target is not None
        ret: List[RemoveObjectAction | CreateObjectAction] = [
            RemoveObjectAction(object_position=target)
        ]
        candidate_list = self._get_candidate_list(match)
        for i in range(2):
            random_target = candidate_list[int(match._random() * len(candidate_list))]
            position = ObjectPosition(
                player_idx=self.position.player_idx, area=ObjectPositionType.HAND, id=-1
            )
            args = {}
            if self._accept_version is not None:
                args["version"] = self._accept_version
            ret.append(
                CreateObjectAction(
                    object_name=random_target,
                    object_position=position,
                    object_arguments=args,
                )
            )
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.TEAM_STATUS),
                object_arguments={},
            )
        )
        return ret


desc: Dict[str, DescDictType] = {
    "TEAM_STATUS/Sunyata Flower": {
        "names": {"en-US": "Sunyata Flower", "zh-CN": "净觉花"},
        "descs": {
            "4.4": {
                "en-US": "During the Round, the next time you play a Support Card: Spend 1 less Elemental Die.",  # noqa: E501
                "zh-CN": "本回合中，我方下次打出支援牌时：少花费1个元素骰。",  # noqa: E501
            }
        },
    },
    "CARD/Sunyata Flower": {
        "names": {"en-US": "Sunyata Flower", "zh-CN": "净觉花"},
        "descs": {
            "4.4": {
                "en-US": "Select a card from your Support Zone and discard it. After that, randomly generate 2 Support Cards in your hand.\nDuring the Round, the next time you play a Support Card: Spend 1 less Elemental Die.",  # noqa: E501
                "zh-CN": "选择一张我方支援区的牌，将其弃置。然后，在我方手牌中随机生成2张支援牌。\n本回合中，我方下次打出支援牌时：少花费1个元素骰。",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Ganluhua.png",  # noqa: E501
        "id": 332029,
    },
}


register_class(SunyataFlower_4_4 | SunyataFlowerStatus_4_4, desc)



from typing import List, Literal

from ...consts import ELEMENT_TO_DAMAGE_TYPE

from ....utils.class_registry import register_class

from ..character_base import ElementalNormalAttackBase
from .rhodeia_3_3 import (
    RhodeiaOfLoch_3_3,
    RhodeiaElementSkill,
    TideAndTorrent as TAT_3_3,
)


class TideAndTorrent(TAT_3_3):
    damage: int = 4
    damage_per_summon: int = 1


class RhodeiaOfLoch_4_2(RhodeiaOfLoch_3_3):
    version: Literal["4.2"] = "4.2"
    skills: List[ElementalNormalAttackBase | RhodeiaElementSkill | TideAndTorrent] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Surge",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            RhodeiaElementSkill(
                name="Oceanid Mimic Summoning",
            ),
            RhodeiaElementSkill(
                name="The Myriad Wilds",
            ),
            TideAndTorrent(),
        ]


register_class(RhodeiaOfLoch_4_2)



from enum import Enum
from ..utils import BaseModel
from typing import Literal, List, Tuple
from .interaction import (
    ChooseCharacterResponse,
    RerollDiceResponse,
    DeclareRoundEndResponse,
    SwitchCharacterResponse,
)
from .consts import DieColor, ObjectType, SkillType
from .modifiable_values import DamageValue
from .struct import MultipleObjectPosition, ObjectPosition


class ActionTypes(str, Enum):
    EMPTY = "EMPTY"
    DRAW_CARD = "DRAW_CARD"
    RESTORE_CARD = "RESTORE_CARD"
    REMOVE_CARD = "REMOVE_CARD"
    CHOOSE_CHARACTER = "CHOOSE_CHARACTER"
    CREATE_DICE = "CREATE_DICE"
    REMOVE_DICE = "REMOVE_DICE"
    DECLARE_ROUND_END = "DECLARE_ROUND_END"
    ACTION_END = "ACTION_END"
    SWITCH_CHARACTER = "SWITCH_CHARACTER"
    CHARGE = "CHARGE"
    USE_SKILL = "USE_SKILL"
    USE_CARD = "USE_CARD"
    SKILL_END = "SKILL_END"
    CHARACTER_DEFEATED = "CHARACTER_DEFEATED"
    CHARACTER_REVIVE = "CHARACTER_REVIVE"
    CREATE_OBJECT = "CREATE_OBJECT"
    CREATE_RANDOM_OBJECT = "CREATE_RANDOM_OBJECT"
    REMOVE_OBJECT = "REMOVE_OBJECT"
    OBJECT_REMOVED = "OBJECT_REMOVED"
    CHANGE_OBJECT_USAGE = "CHANGE_OBJECT_USAGE"
    MOVE_OBJECT = "MOVE_OBJECT"
    CONSUME_ARCANE_LEGEND = "CONSUME_ARCANE_LEGEND"

    # system phase events
    GAME_START = "GAME_START"
    ROUND_PREPARE = "ROUND_PREPARE"
    PLAYER_ACTION_START = "PLAYER_ACTION_START"
    ROUND_END = "ROUND_END"

    # make damage related events
    RECEIVE_DAMAGE = "RECEIVE_DAMAGE"
    MAKE_DAMAGE = "MAKE_DAMAGE"
    AFTER_MAKE_DAMAGE = "AFTER_MAKE_DAMAGE"

    # generate request actions
    GENERATE_CHOOSE_CHARACTER = "GENERATE_CHOOSE_CHARACTER"
    GENERATE_REROLL_DICE = "GENERATE_REROLL_DICE"
    GENERATE_SWITCH_CARD = "GENERATE_SWITCH_CARD"

    # change game state actions
    SKIP_PLAYER_ACTION = "SKIP_PLAYER_ACTION"


class ActionBase(BaseModel):
    """
    Base class for game actions.
    An action contains arguments to make changes to the game table.

    Attributes:
        action_type (Literal[ActionTypes]): The type of the action.
        record_level (int): The level of the action to record in match.history,
            lower level means more important. It should be a positive number.
    """

    type: Literal[ActionTypes.EMPTY] = ActionTypes.EMPTY

    record_level: int = 100


class DrawCardAction(ActionBase):
    """
    Action for drawing cards.
    """

    type: Literal[ActionTypes.DRAW_CARD] = ActionTypes.DRAW_CARD
    record_level: int = 10
    player_idx: int
    number: int
    blacklist_names: List[str] = []
    whitelist_names: List[str] = []
    blacklist_types: List[ObjectType] = []
    whitelist_types: List[ObjectType] = []
    blacklist_cost_labels: int = 0
    whitelist_cost_labels: int = 0
    draw_if_filtered_not_enough: bool


class RestoreCardAction(ActionBase):
    """
    Action for restoring cards.
    """

    type: Literal[ActionTypes.RESTORE_CARD] = ActionTypes.RESTORE_CARD
    player_idx: int
    card_idxs: List[int]


class RemoveCardAction(ActionBase):
    """
    Action for removing cards.
    """

    type: Literal[ActionTypes.REMOVE_CARD] = ActionTypes.REMOVE_CARD
    record_level: int = 20
    position: ObjectPosition
    remove_type: Literal["USED", "BURNED"]


class ChooseCharacterAction(ActionBase):
    """
    Action for choosing characters.
    """

    type: Literal[ActionTypes.CHOOSE_CHARACTER] = ActionTypes.CHOOSE_CHARACTER
    record_level: int = 10
    player_idx: int
    character_idx: int

    @classmethod
    def from_response(cls, response: ChooseCharacterResponse):
        """
        Generate ChooseCharacterAction from ChooseCharacterResponse.
        """
        return cls(
            player_idx=response.player_idx,
            character_idx=response.character_idx,
        )


# 5


class CreateDiceAction(ActionBase):
    """
    Action for creating dice.

    Args:
        player_idx (int): The index of the player to create the dice for.
        number (int): The number of dice to create.
        color (DieColor | None): The color of the dice to create. If None,
            the following generate rules will be activated.
        random (bool): Whether to randomly generate the color of dice.
        different (bool): Whether to generate different colors of dice.
            if you want to generate random but never OMNI die, generate one
            different die. If multiple, use multiple action.
    """

    type: Literal[ActionTypes.CREATE_DICE] = ActionTypes.CREATE_DICE
    player_idx: int
    number: int
    color: DieColor | None = None
    random: bool = False
    different: bool = False


class RemoveDiceAction(ActionBase):
    """
    Action for removing dice.

    Args:
        player_idx (int): The index of the player to remove the dice for.
        dice_idxs (List[int]): The indices of the dice to remove.
    """

    type: Literal[ActionTypes.REMOVE_DICE] = ActionTypes.REMOVE_DICE
    player_idx: int
    dice_idxs: List[int]

    @classmethod
    def from_response(cls, response: RerollDiceResponse):
        """
        Generate RemoveDiceAction from RerollDiceResponse.
        """
        return cls(
            player_idx=response.player_idx,
            dice_idxs=response.reroll_dice_idxs,
        )


class DeclareRoundEndAction(ActionBase):
    """
    Action for declaring the end of the round.
    """

    type: Literal[ActionTypes.DECLARE_ROUND_END] = ActionTypes.DECLARE_ROUND_END
    record_level: int = 10
    player_idx: int

    @classmethod
    def from_response(cls, response: DeclareRoundEndResponse):
        """
        Generate DeclareRoundEndAction from DeclareRoundEndResponse.
        """
        return cls(
            player_idx=response.player_idx,
        )


class ActionEndAction(ActionBase):
    """
    Action end, and if is a combat action, change current player.
    the position means the action source, i.e. the skill,
    or the character who switch out.
    """

    type: Literal[ActionTypes.ACTION_END] = ActionTypes.ACTION_END
    action_label: int  # Refer to PlayerActionLabels
    do_combat_action: bool
    position: ObjectPosition


class SwitchCharacterAction(ActionBase):
    """
    Action for switching character.
    """

    type: Literal[ActionTypes.SWITCH_CHARACTER] = ActionTypes.SWITCH_CHARACTER
    record_level: int = 10
    player_idx: int
    character_idx: int

    @classmethod
    def from_response(cls, response: SwitchCharacterResponse):
        """
        Generate SwitchCharacterAction from SwitchCharacterResponse.
        """
        return cls(
            player_idx=response.player_idx,
            character_idx=response.request.target_character_idx,
        )


# 10


class ChargeAction(ActionBase):
    """
    Action for charging.
    """

    type: Literal[ActionTypes.CHARGE] = ActionTypes.CHARGE
    player_idx: int
    character_idx: int
    charge: int


class UseSkillAction(ActionBase):
    """
    Action for using skill.
    """

    type: Literal[ActionTypes.USE_SKILL] = ActionTypes.USE_SKILL
    record_level: int = 10
    skill_position: ObjectPosition


class UseCardAction(ActionBase):
    """
    Action for using card.
    """

    type: Literal[ActionTypes.USE_CARD] = ActionTypes.USE_CARD
    record_level: int = 10
    card_position: ObjectPosition
    target: ObjectPosition | MultipleObjectPosition | None


class SkillEndAction(ActionBase):
    """
    Action for ending skill.
    """

    type: Literal[ActionTypes.SKILL_END] = ActionTypes.SKILL_END
    position: ObjectPosition
    target_position: ObjectPosition  # always poopnent initial active character now  # noqa: E501
    skill_type: SkillType


class CharacterDefeatedAction(ActionBase):
    """
    Action for character defeated.
    """

    type: Literal[ActionTypes.CHARACTER_DEFEATED] = ActionTypes.CHARACTER_DEFEATED
    player_idx: int
    character_idx: int


# 15


class CreateObjectAction(ActionBase):
    """
    Action for creating objects, e.g. status, summons, supports.
    Note some objects are not created but moved, e.g. equipment and supports,
    for these objects, do not use this action.
    """

    type: Literal[ActionTypes.CREATE_OBJECT] = ActionTypes.CREATE_OBJECT
    object_position: ObjectPosition
    object_name: str
    object_arguments: dict


class CreateRandomObjectAction(ActionBase):
    """
    Action for creating random objects. A list of names are provided, and
    `number` of them will be created by uniformly selecting from the list and doing
    `CreateObjectAction` for each of them.
    NOTE: some actions will not generate existing objects, e.g. elemental skill of
    Rhodeia will not summoning existing objects unless full. This action will not deal
    with such situation, you need to write logic to filter out existing objects before
    creating this action.
    """

    type: Literal[ActionTypes.CREATE_OBJECT] = ActionTypes.CREATE_OBJECT
    object_position: ObjectPosition
    object_names: List[str]
    object_arguments: dict
    number: int

    def select_by_idx(
        self, idx: int
    ) -> Tuple[CreateObjectAction, "CreateRandomObjectAction"]:
        """
        Select an object name by index and create a `CreateObjectAction` for it.
        It will also return a new `CreateRandomObjectAction` with the same
        arguments but with the `number` reduced by 1 and the selected object
        name removed from the list.

        Returns:
            CreateObjectAction: The action for creating the object.
            CreateRandomObjectAction: The new action with the `number` reduced
                by 1 and the selected object name removed from the list.
        """
        selected = self.object_names[idx]
        others = self.object_names[:idx] + self.object_names[idx + 1 :]
        return (
            CreateObjectAction(
                object_position=self.object_position,
                object_name=selected,
                object_arguments=self.object_arguments,
            ),
            CreateRandomObjectAction(
                object_position=self.object_position,
                object_names=others,
                object_arguments=self.object_arguments,
                number=self.number - 1,
            ),
        )


class RemoveObjectAction(ActionBase):
    """
    Action for removing objects.

    Args:
        object_position (ObjectPosition): The ID and position of the
            object to remove.
    """

    type: Literal[ActionTypes.REMOVE_OBJECT] = ActionTypes.REMOVE_OBJECT
    object_position: ObjectPosition


class ChangeObjectUsageAction(ActionBase):
    """
    Action for changing object usage. change in delta.
    """

    type: Literal[ActionTypes.CHANGE_OBJECT_USAGE] = ActionTypes.CHANGE_OBJECT_USAGE
    object_position: ObjectPosition
    change_usage: int
    min_usage: int = 0
    max_usage: int = 999


class MoveObjectAction(ActionBase):
    """
    Action for moving objects.
    """

    type: Literal[ActionTypes.MOVE_OBJECT] = ActionTypes.MOVE_OBJECT
    record_level: int = 10
    object_position: ObjectPosition
    target_position: ObjectPosition
    # for Master of Weaponry etc. in 4.1, when mark true, reset round usage
    reset_usage: bool = False


# 20


class MakeDamageAction(ActionBase):
    """
    Action for making damage. Heal treats as negative damage. Elemental
    applies to the character (e.g. Kokomi) treats as zero damage.

    It can also contain character change and object creation (caused by
    elemental reaction or skill effects). They will be executed right after
    making damage, and gives corresponding events.

    Args:
        damage_value_list (List[DamageValue]): The damage values to make.
        do_character_change (bool): Whether to change character after making
            damage.
        character_change_idx (List[int]): The character indices of the
            character who will be changed to for each player. If it is -1,
            this damage will not explicitly change the character. It should
            not be a defeated character.
        create_objects (List[CreateObjectAction]): The objects to create after
            making damage.
    """

    type: Literal[ActionTypes.MAKE_DAMAGE] = ActionTypes.MAKE_DAMAGE
    record_level: int = 10
    damage_value_list: List[DamageValue]
    create_objects: List[CreateObjectAction] = []

    # character change
    character_change_idx: List[int] = [-1, -1]

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        for damage_value in self.damage_value_list:
            assert (
                damage_value.position.id == self.damage_value_list[0].position.id
            ), "all damage should from same source"


class ConsumeArcaneLegendAction(ActionBase):
    """
    Action for consuming arcane legend.
    """

    type: Literal[ActionTypes.CONSUME_ARCANE_LEGEND] = ActionTypes.CONSUME_ARCANE_LEGEND
    player_idx: int


class GenerateChooseCharacterRequestAction(ActionBase):
    """
    Action for generating choose character action.
    """

    type: Literal[
        ActionTypes.GENERATE_CHOOSE_CHARACTER
    ] = ActionTypes.GENERATE_CHOOSE_CHARACTER
    player_idx: int


class GenerateRerollDiceRequestAction(ActionBase):
    """
    Action for generating reroll dice request.
    """

    type: Literal[ActionTypes.GENERATE_REROLL_DICE] = ActionTypes.GENERATE_REROLL_DICE
    player_idx: int
    reroll_times: int


class SkipPlayerActionAction(ActionBase):
    """
    Action for skipping current player action.
    """

    type: Literal[ActionTypes.SKIP_PLAYER_ACTION] = ActionTypes.SKIP_PLAYER_ACTION


# 25


class CharacterReviveAction(ActionBase):
    """
    Action for character revive.
    """

    type: Literal[ActionTypes.CHARACTER_REVIVE] = ActionTypes.CHARACTER_REVIVE
    record_level: int = 10
    player_idx: int
    character_idx: int
    revive_hp: int


class GenerateSwitchCardRequestAction(ActionBase):
    """
    Action for generating switch card request.
    """

    type: Literal[ActionTypes.GENERATE_SWITCH_CARD] = ActionTypes.GENERATE_SWITCH_CARD
    player_idx: int


Actions = (
    ActionBase
    | DrawCardAction
    | RestoreCardAction
    | RemoveCardAction
    | ChooseCharacterAction
    # 5
    | CreateDiceAction
    | RemoveDiceAction
    | DeclareRoundEndAction
    | ActionEndAction
    | SwitchCharacterAction
    # 10
    | ChargeAction
    | UseSkillAction
    | UseCardAction
    | SkillEndAction
    | CharacterDefeatedAction
    # 15
    | CreateObjectAction
    | CreateRandomObjectAction
    | RemoveObjectAction
    | ChangeObjectUsageAction
    | MoveObjectAction
    # 20
    | MakeDamageAction
    | ConsumeArcaneLegendAction
    | GenerateChooseCharacterRequestAction
    | GenerateRerollDiceRequestAction
    | SkipPlayerActionAction
    # 25
    | CharacterReviveAction
    | GenerateSwitchCardRequestAction
)



from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import IconType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundTeamStatus, ShieldTeamStatus


class RebelliousShield_3_7(ShieldTeamStatus):
    name: Literal["Rebellious Shield"] = "Rebellious Shield"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 2


class MillennialMovementFarewellSong_3_7(RoundTeamStatus):
    name: Literal[
        "Millennial Movement: Farewell Song"
    ] = "Millennial Movement: Farewell Song"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            assert mode == "REAL"
            value.damage += 1
        return value


register_class(RebelliousShield_3_7 | MillennialMovementFarewellSong_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...consts import ELEMENT_TO_DAMAGE_TYPE, DieColor

from ...struct import Cost
from ...action import Actions
from ..character_base import ElementalNormalAttackBase, SkillTalent
from .sangonomiya_kokomi_3_6 import SangonomiyaKokomi_3_6 as SK_3_6
from .sangonomiya_kokomi_3_6 import NereidsAscension as NA_3_6
from .sangonomiya_kokomi_3_6 import KuragesOath


class NereidsAscension(NA_3_6):
    damage: int = 3

    def get_actions(self, match: Any) -> List[Actions]:
        """
        No healing
        """
        return super(NA_3_6, self).get_actions(
            match,
            [
                self.create_character_status("Ceremonial Garment"),
            ],
        )


class TamakushiCasket_3_5(SkillTalent):
    name: Literal["Tamakushi Casket"]
    version: Literal["3.5"] = "3.5"
    character_name: Literal["Sangonomiya Kokomi"] = "Sangonomiya Kokomi"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )
    skill: Literal["Nereid's Ascension"] = "Nereid's Ascension"


class SangonomiyaKokomi_3_5(SK_3_6):
    version: Literal["3.5"]
    skills: List[ElementalNormalAttackBase | KuragesOath | NereidsAscension] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="The Shape of Water",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            KuragesOath(),
            NereidsAscension(),
        ]


register_class(SangonomiyaKokomi_3_5 | TamakushiCasket_3_5)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue

from ...summon.base import AttackerSummonBase

from ...action import Actions, ChargeAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class EyeOfStormyJudgment_3_7(AttackerSummonBase):
    name: Literal["Eye of Stormy Judgment"] = "Eye of Stormy Judgment"
    version: Literal["3.7"] = "3.7"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if (
            not value.is_corresponding_character_use_damage_skill(
                self.position, match, SkillType.ELEMENTAL_BURST
            )
            or value.damage_elemental_type == DamageElementalType.PIERCING
            or value.damage_from_element_reaction
        ):
            # not corresponding character use elemental burst, or is piercing
            # damage, or damage is from element reaction
            return value
        # add damage
        value.damage += 1
        return value


# Skills


class TranscendenceBalefulOmen(ElementalSkillBase):
    name: Literal["Transcendence: Baleful Omen"] = "Transcendence: Baleful Omen"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [self.charge_self(1), self.create_summon("Eye of Stormy Judgment")]


class SecretArtMusouShinsetsu(ElementalBurstBase):
    name: Literal["Secret Art: Musou Shinsetsu"] = "Secret Art: Musou Shinsetsu"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        table = match.player_tables[self.position.player_idx]
        for cid, character in enumerate(table.characters):
            if character.is_alive and cid != self.position.character_idx:
                ret.append(
                    ChargeAction(
                        player_idx=self.position.player_idx,
                        character_idx=cid,
                        charge=2,
                    )
                )
        return ret


class ChakraDesiderata(CreateStatusPassiveSkill):
    name: Literal["Chakra Desiderata"] = "Chakra Desiderata"
    status_name: Literal["Chakra Desiderata"] = "Chakra Desiderata"


# Talents


class WishesUnnumbered_3_7(SkillTalent):
    name: Literal["Wishes Unnumbered"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Raiden Shogun"] = "Raiden Shogun"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Secret Art: Musou Shinsetsu"] = "Secret Art: Musou Shinsetsu"


# character base


class RaidenShogun_3_7(CharacterBase):
    name: Literal["Raiden Shogun"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | TranscendenceBalefulOmen
        | SecretArtMusouShinsetsu
        | ChakraDesiderata
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Origin",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TranscendenceBalefulOmen(),
            SecretArtMusouShinsetsu(),
            ChakraDesiderata(),
        ]


register_class(RaidenShogun_3_7 | WishesUnnumbered_3_7 | EyeOfStormyJudgment_3_7)



from typing import Dict, List, Literal
from pydantic import PrivateAttr

from ....action import DrawCardAction
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType
from ....match import Match

from ....consts import CostLabels
from ....card.support.companions import Timaeus_3_3


class Timaeus_4_3(Timaeus_3_3):
    version: Literal["4.3"] = "4.3"
    decrease_target: int = CostLabels.ARTIFACT.value
    _check_type: Literal["TYPE", "NUMBER"] = PrivateAttr("NUMBER")
    _check_number: int = PrivateAttr(6)

    def play(self, match: Match) -> List[DrawCardAction]:
        if self._check_type == "NUMBER":
            counter = 0
            for card in match.player_tables[
                self.position.player_idx
            ].player_deck_information.cards:
                if card.cost_label & self.decrease_target != 0:
                    counter += 1
            if counter >= self._check_number:
                return [
                    DrawCardAction(
                        player_idx=self.position.player_idx,
                        number=1,
                        whitelist_cost_labels=self.decrease_target,
                        draw_if_filtered_not_enough=False,
                    )
                ]
        else:
            assert self._check_type == "TYPE"
            names = set()
            for card in match.player_tables[
                self.position.player_idx
            ].player_deck_information.cards:
                if card.cost_label & self.decrease_target != 0:
                    names.add(card.name)
            if len(names) >= self._check_number:
                return [
                    DrawCardAction(
                        player_idx=self.position.player_idx,
                        number=1,
                        whitelist_cost_labels=self.decrease_target,
                        draw_if_filtered_not_enough=False,
                    )
                ]
        return []


class Wagner_4_3(Timaeus_4_3):
    name: Literal["Wagner"]
    decrease_target: int = CostLabels.WEAPON.value
    _check_type: Literal["TYPE", "NUMBER"] = PrivateAttr("TYPE")
    _check_number: int = PrivateAttr(3)


desc: Dict[str, DescDictType] = {
    "SUPPORT/Wagner": {
        "descs": {
            "4.3": {
                "zh-CN": "入场时附带2个「锻造原胚」。结束阶段：补充1个「锻造原胚」。入场时：如果牌组中携带了至少3种「武器」，从牌组中随机抽取1张「武器」。打出「武器」手牌时：如可能，则支付等同于「武器」总费用数量的「锻造原胚」，以免费装备此「武器」。（每回合1次）",  # noqa: E501
                "en-US": "Comes with 2 Forging Billet when played. End Phase: Gain 1 Forging Billet. When played: If you have not less than 3 types of Weapons in your initial deck, draw 1 Weapon from your deck. When playing a Weapon Card: If possible, spend Forging Billet equal to the total cost of the Weapon and equip this Weapon for free. (Once per Round)",  # noqa: E501
            }
        },
    },
    "SUPPORT/Timaeus": {
        "descs": {
            "4.3": {
                "zh-CN": "入场时附带2个「合成材料」。结束阶段：补充1个「合成材料」。入场时：如果牌组中携带了至少6张「圣遗物」，从牌组中随机抽取1张「圣遗物」。打出「圣遗物」手牌时：如可能，则支付等同于「圣遗物」总费用数量的「合成材料」，以免费装备此「圣遗物」。（每回合1次）",  # noqa: E501
                "en-US": "Comes with 2 Transmutation Materials when played. When played: If you have not less than 6 Artifacts in your initial deck, draw 1 Artifact from your deck. End Phase: Gain 1 Transmutation Material. When playing an Artifact Card: If possible, spend Transmutation Materials equal to the total cost of the Artifact and equip this Artifact for free. (Once per Round)",  # noqa: E501
            }
        },
    },
}


register_class(Timaeus_4_3 | Wagner_4_3, desc)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....consts import ELEMENT_TO_DIE_COLOR, DieColor
from ....match import Match
from ....action import CreateDiceAction
from ....struct import Cost
from ....card.equipment.artifact.gilded_dreams import ShadowOfTheSandKing_4_2
from .....utils.desc_registry import DescDictType


class GildedDreams_4_3(ShadowOfTheSandKing_4_2):
    name: Literal["Gilded Dreams"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(any_dice_number=3)
    max_usage_per_round: int = 2

    def equip(self, match: Match) -> List[CreateDiceAction]:
        """
        When equipped, create one elemental dice. If there are 3 different
        elemental dice in your team, create one omni dice.
        """
        super().equip(match)
        characters = match.player_tables[self.position.player_idx].characters
        equip_character = characters[self.position.character_idx]
        ret: List[CreateDiceAction] = [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=ELEMENT_TO_DIE_COLOR[equip_character.element],
            )
        ]
        elements = set()
        for character in characters:
            elements.add(character.element)
        if len(elements) >= 3:
            ret.append(
                CreateDiceAction(
                    player_idx=self.position.player_idx, number=1, color=DieColor.OMNI
                )
            )
        return ret


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Gilded Dreams": {
        "names": {"en-US": "Gilded Dreams", "zh-CN": "饰金之梦"},
        "descs": {
            "4.3": {
                "en-US": "When played: generate 1 Die of the same Element as the attached character. If you have 3 different Elemental Types in your party, generate 1 Omni Element in addition to this.\nWhen the character to which this card is attached is your active character, then when an opposing character takes Elemental Reaction DMG: Draw a card. (Twice per Round)\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "入场时：生成1个所附属角色类型的元素骰。如果我方队伍中存在3种不同元素类型的角色，则额外生成1个万能元素。\n所附属角色为出战角色期间，敌方受到元素反应伤害时：抓1张牌。（每回合至多2次）\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_Chiwangtao.png",  # noqa: E501
        "id": 312018,
    },
}


register_class(GildedDreams_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase

from ...modifiable_values import DamageIncreaseValue
from ...event import (
    CharacterReviveEventArguments,
    ReceiveDamageEventArguments,
    RoundPrepareEventArguments,
)

from ...action import Actions, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class Ushi_3_6(DefendSummonBase):
    name: Literal["Ushi"] = "Ushi"
    version: Literal["3.6"] = "3.6"
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = False

    create_status_triggered: bool = False

    def renew(self, new_status: "Ushi_3_6") -> None:
        self.create_status_triggered = new_status.create_status_triggered
        return super().renew(new_status)

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If any of our character is attacked and not triggered, Arataki Itto
        gains Superlative Superstrength.
        """
        if self.create_status_triggered:
            # already triggered, no effect
            return []
        if event.final_damage.target_position.player_idx != self.position.player_idx:
            # not attack our player, no effect
            return []
        if event.final_damage.damage_type != DamageType.DAMAGE:
            # not damage, no effect
            return []
        # trigger, gain Superlative Superstrength for first Arataki Itto
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.is_alive and character.name == "Arataki Itto":
                # found alive itto, gain status
                status_position = character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                )
                self.create_status_triggered = True
                return [
                    CreateObjectAction(
                        object_name="Superlative Superstrength",
                        object_position=status_position,
                        object_arguments={},
                    )
                ]
        else:  # pragma: no cover
            # not found alive Itto, no effect
            return []


# Skills


class FightClubLegend(PhysicalNormalAttackBase):
    name: Literal["Fight Club Legend"] = "Fight Club Legend"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=1,
        any_dice_number=2,
    )


class MasatsuZetsugiAkaushiBurst(ElementalSkillBase):
    name: Literal["Masatsu Zetsugi: Akaushi Burst!"] = "Masatsu Zetsugi: Akaushi Burst!"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(elemental_dice_color=DieColor.GEO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, create Ushi and create status
        """
        ret = super().get_actions(
            match,
            [
                self.create_summon("Ushi"),
                self.create_character_status("Superlative Superstrength"),
            ],
        )
        return ret


class RoyalDescentBeholdIttoTheEvil(ElementalBurstBase):
    name: Literal[
        "Royal Descent: Behold, Itto the Evil!"
    ] = "Royal Descent: Behold, Itto the Evil!"
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=3
    )
    version: Literal["3.6"] = "3.6"

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create status
        """
        return super().get_actions(
            match,
            [
                self.create_character_status(
                    "Raging Oni King", {"version": self.version}
                )
            ],
        )


# Talents


class AratakiIchiban_3_6(SkillTalent):
    name: Literal["Arataki Ichiban"]
    version: Literal["3.6"] = "3.6"
    character_name: Literal["Arataki Itto"] = "Arataki Itto"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO,
        elemental_dice_number=1,
        any_dice_number=2,
    )
    skill: Literal["Fight Club Legend"] = "Fight Club Legend"


# character base


class AratakiItto_3_6(CharacterBase):
    name: Literal["Arataki Itto"]
    version: Literal["3.6"] = "3.6"
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        FightClubLegend | MasatsuZetsugiAkaushiBurst | RoyalDescentBeholdIttoTheEvil
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    normal_attack_this_round: int = 0

    def _init_skills(self) -> None:
        self.skills = [
            FightClubLegend(),
            MasatsuZetsugiAkaushiBurst(),
            RoyalDescentBeholdIttoTheEvil(),
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        # reset normal attack counter
        self.normal_attack_this_round = 0
        return []

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[Actions]:
        """
        When self is revived, clear counter
        """
        if (
            event.action.player_idx == self.position.player_idx
            and event.action.character_idx == self.position.character_idx
        ):
            self.normal_attack_this_round = 0
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ):
        """
        When use normal attack, add counter.

        For the second time or more using normal attack this round,
        and is charged attack, and has status Superlative Superstrength,
        and equipped talent, add 1 damage.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not use normal attack
            return value
        # add counter
        self.normal_attack_this_round += 1
        if self.normal_attack_this_round <= 1:
            # first time, no effect
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, no effect
            return value
        status_found = False
        for status in self.status:
            if status.name == "Superlative Superstrength":
                status_found = True
                break
        if not status_found:
            # not have status, no effect
            return value
        if self.talent is None:
            # not equipped talent, no effect
            return value
        # add damage
        value.damage += 1
        return value


register_class(AratakiItto_3_6 | AratakiIchiban_3_6 | Ushi_3_6)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import CreateObjectEventArguments
from ...action import Actions, ChangeObjectUsageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class Prowl(ElementalSkillBase):
    name: Literal["Prowl"] = "Prowl"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match, [self.create_character_status("Stealth")])


class StealthMaster(CreateStatusPassiveSkill):
    name: Literal["Stealth Master"] = "Stealth Master"
    status_name: Literal["Stealth"] = "Stealth"
    regenerate_when_revive: bool = False


# Talents


class PaidinFull_3_3(SkillTalent):
    name: Literal["Paid in Full"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Fatui Pyro Agent"] = "Fatui Pyro Agent"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    skill: Literal["Prowl"] = "Prowl"

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When create object, check if it is stealth
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.object_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.CHARACTER_STATUS,
        ):
            # not self create character status, do nothing
            return []
        if event.action.object_name != "Stealth":
            # not stealth, do nothing
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        status = None
        for s in character.status:
            if s.name == "Stealth":
                status = s
                break
        else:
            raise AssertionError("Stealth not found")
        # add 1 usage
        return [
            ChangeObjectUsageAction(object_position=status.position, change_usage=1)
        ]


# character base


class FatuiPyroAgent_3_3(CharacterBase):
    name: Literal["Fatui Pyro Agent"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Prowl | ElementalBurstBase | StealthMaster
    ] = []
    faction: List[FactionType] = [FactionType.FATUI]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Thrust",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Prowl(),
            ElementalBurstBase(
                name="Blade Ablaze",
                damage=5,
                damage_type=DamageElementalType.PYRO,
                cost=ElementalBurstBase.get_cost(self.element, 3, 2),
            ),
            StealthMaster(),
        ]


register_class(FatuiPyroAgent_3_3 | PaidinFull_3_3)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_keqing():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "end",
            "skill 0 0 1 2",
            "card 0 0 0 1 2",
            "skill 0 0 1 2",
            "TEST 1 20 20 20 14 20 13",
            "card 6 0 0 1 2",
            "skill 2 0 1 2 3",
            "TEST 1 20 20 20 11 17 5",
            "end",
            "end",
            "sw_char 1 0",
            "end",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 1 6 3 18 11 15 5",
            "TEST 3 p1c1 ele app",
            "skill 1 0 1 2",
            "TEST 1 6 3 18 11 12 5",
            "TEST 4 hand 10 10",
            "skill 1 0 1 2",
            "TEST 2 p0c0 status usage",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
            "sw_char 2 0",
            "TEST 1 20 20 20 14 20 18",
            "TEST 3 p1c2 ele application electro",
            "end",
            "sw_char 0 0",
            "skill 1 0 1 2",
            "card 0 0 0 1 2",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "end",
            "skill 0 0 1 2",
            "skill 2 0 1 2 3",
            "card 8 0 0 1 2",
            "TEST 1 7 10 19 11 17 5",
            "TEST 3 p0c1 ele app",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Keqing
        character:Kaeya
        character:Fischl
        Thundering Penance*30
        """
    )
    for character in deck.characters:
        character.hp = 20
        character.max_hp = 20
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                cid = int(cmd[2][3])
                usage = [int(x) for x in cmd[5:]]
                status = match.player_tables[pid].characters[cid].status
                assert len(status) == len(usage)
            elif test_id == 3:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                cid = int(cmd[2][3])
                app = [x.upper() for x in cmd[5:]]
                character = match.player_tables[pid].characters[cid]
                assert character.element_application == app
            elif test_id == 4:
                hand = [int(x) for x in cmd.split()[3:]]
                assert len(hand) == 2
                for table, h in zip(match.player_tables, hand):
                    assert len(table.hands) == h
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_keqing_2():
    """
    This test mainly tests the following:
        1. when hands have multiple Lightning Stiletto, use skill will remove
            all of them (to get multiple Lightning Stiletto, use Nature and
            Wisdom to put them into deck)
        2. Artifacts can decrease the cost of Lightning Stiletto, even if
            Keqing is not active character
        3. When Keqing defeated, all Lightning Stiletto cannot use.
    """
    cmd_records = [
        [
            "sw_card 1 3",
            "choose 1",
            "card 0 1 15 14",
            "skill 1 13 12",
            "sw_char 0 11",
            "end",
            "card 4 0 15 14",
            "sw_char 2 13",
            "card 3 0 12",
            "sw_card 0 1 2 4 5 3",
            "sw_char 1 11",
            "skill 1 10 9 8",
            "card 1 0 7",
            "sw_card 5 3 2 1 0",
            "end",
            "sw_char 2 15",
            "sw_char 0 14",
            "sw_char 1 13",
            "skill 1 12 11",
            "card 7 0 10",
            "sw_card 7 8 9",
            "card 6 2 9 8",
            "card 7 0 7",
            "sw_card 6 7",
            "end",
            "card 7 0 15",
            "sw_card 8 7 6 5 4 3 2 1 0",
            "end",
            "choose 2",
            "TEST 3 card 0 1 cannot use",
            "end",
        ],
        [
            "sw_card 1",
            "choose 1",
            "skill 1 15 14 13",
            "card 4 1 12 11",
            "TEST 2 card 4 cost 2",
            "card 3 0 10",
            "sw_card 3",
            "skill 1 9 8",
            "card 4 0 7",
            "sw_card 4 5",
            "skill 1 6 5 4",
            "card 5 0 3",
            "sw_card 5",
            "skill 1 2 1 0",
            "end",
            "end",
            "card 5 0 15",
            "sw_card 5 6 7 9",
            "skill 1 14 13",
            "card 9 0 12",
            "sw_card",
            "card 9 2 11 10",
            "skill 1 9 8 7",
            "sw_char 0 6",
            "card 8 0 5",
            "sw_card 8 7",
            "card 8 2 4 3",
            "card 8 0 2 1",
            "end",
            "sw_char 1 15",
            "card 8 0 14",
            "sw_card 7 6 9 5 4 3 2 1 0",
            "card 3 0 13 12",
            "TEST 1 p1 hand 5",
            "skill 2 11 10 9 8",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Kaeya
        character:Keqing
        character:Fischl
        Nature and Wisdom*15
        Thunder Summoner's Crown*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                assert len(match.player_tables[1].hands) == 5
            elif test_id == 2:
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == 4:
                        assert req.cost.total_dice_cost == 2
            elif test_id == 3:
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx not in [0, 1]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_keqing_freeze():
    """
    If keqing is frozen, it cannot use Lightning Stiletto.
    """
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "sw_char 1 11",
            "sw_char 2 10",
            "skill 1 9 8 7",
            "end",
            "sw_char 0 15",
            "skill 1 14 13 12",
        ],
        [
            "sw_card 1 2",
            "choose 1",
            "skill 1 15 14 13",
            "TEST 1 card 5 can use",
            "sw_char 0 12",
            "TEST 1 card 5 can use",
            "sw_char 1 11",
            "TEST 2 card 5 cannot use",
            "sw_char 0 10",
            "TEST 2 card 5 cannot use",
            "end",
            "TEST 1 card 5 can use",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Kaeya
        character:Keqing
        character:Mona
        Nature and Wisdom*15
        Thunder Summoner's Crown*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(test_cmd.split()[3])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        break
                else:
                    raise AssertionError("Request not found")
            elif test_id == 2:
                cidx = int(test_cmd.split()[3])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != cidx
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_keqing()
    test_keqing_2()
    # test_keqing_freeze()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import ActionTypes, Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class YueguiThrowingMode_4_1(AttackerSummonBase):
    name: Literal["Yuegui: Throwing Mode"] = "Yuegui: Throwing Mode"
    version: Literal["4.1"] = "4.1"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1
    is_talent_activated: bool = False

    def renew(self, obj: "YueguiThrowingMode_4_1") -> None:
        self.is_talent_activated = obj.is_talent_activated
        super().renew(obj)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When in round end, attack enemy active character and heal our
        character. When talent activated, and usage is 1, damage is increased
        to 2.
        """
        if self.usage == 1 and self.is_talent_activated:
            self.damage = 2
        ret = super().event_handler_ROUND_END(event, match)
        table = match.player_tables[self.position.player_idx]
        characters = table.characters
        # select character with most damage taken
        selected_character = characters[table.active_character_idx]
        for c in characters:
            if c.is_alive and c.damage_taken > selected_character.damage_taken:
                selected_character = c
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        damage_action.damage_value_list.append(
            DamageValue(
                position=self.position,
                damage_type=DamageType.HEAL,
                target_position=selected_character.position,
                damage=-self.damage,
                damage_elemental_type=DamageElementalType.HEAL,
                cost=Cost(),
            )
        )
        return ret


# Skills


class RaphanusSkyCluster(ElementalSkillBase):
    name: Literal["Raphanus Sky Cluster"] = "Raphanus Sky Cluster"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        args = {}
        if self.is_talent_equipped(match):
            args["is_talent_activated"] = True
        return [self.create_summon("Yuegui: Throwing Mode", args), self.charge_self(1)]


class MoonjadeDescent(ElementalBurstBase):
    name: Literal["Moonjade Descent"] = "Moonjade Descent"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=4, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match, [self.create_team_status("Adeptal Legacy")])


# Talents


class Beneficent_4_1(SkillTalent):
    name: Literal["Beneficent"]
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Yaoyao"] = "Yaoyao"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
    )
    skill: Literal["Raphanus Sky Cluster"] = "Raphanus Sky Cluster"


# character base


class Yaoyao_4_1(CharacterBase):
    name: Literal["Yaoyao"]
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | RaphanusSkyCluster | MoonjadeDescent] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Toss 'N' Turn Spear",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            RaphanusSkyCluster(),
            MoonjadeDescent(),
        ]


register_class(Yaoyao_4_1 | Beneficent_4_1 | YueguiThrowingMode_4_1)



from ...utils import import_all_modules
from .base import StatusBase
from .character_status.base import CharacterStatusBase
from .team_status.base import TeamStatusBase


import_all_modules(__file__, __name__)
__all__ = ("StatusBase", "CharacterStatusBase", "TeamStatusBase")



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundPrepareEventArguments, SkillEndEventArguments

from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DamageType,
    IconType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)

from ...modifiable_values import (
    DamageElementEnhanceValue,
    DamageIncreaseValue,
    DamageValue,
)

from ...action import MakeDamageAction
from .base import (
    DefendTeamStatus,
    RoundTeamStatus,
    ShieldTeamStatus,
    SwitchActionTeamStatus,
    UsageTeamStatus,
    ElementalInfusionTeamStatus,
)


class Icicle_3_3(SwitchActionTeamStatus):
    name: Literal["Icicle"] = "Icicle"
    version: Literal["3.3"] = "3.3"
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def _action(self, match: Any) -> List[MakeDamageAction]:
        """
        attack enemy active character
        """
        active_character = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=active_character.position,
                        damage=2,
                        damage_elemental_type=DamageElementalType.CRYO,
                        cost=Cost(),
                    )
                ]
            )
        ]


class IcyQuill_3_7(UsageTeamStatus):
    # TODO: has talent effect
    name: Literal["Icy Quill"] = "Icy Quill"
    version: Literal["3.7"] = "3.7"
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.ATK_UP_ICE] = IconType.ATK_UP_ICE

    talent_usage: int = 0
    talent_max_usage: int = 0

    def renew(self, new_status: "IcyQuill_3_7") -> None:
        super().renew(new_status)
        self.talent_max_usage = max(new_status.talent_max_usage, self.talent_max_usage)
        self.talent_usage = max(new_status.talent_usage, self.talent_usage)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ):
        """
        When in round prepare, reset talent usage
        """
        self.talent_usage = self.talent_max_usage
        return []

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        """
        If our character skill or summon deal corresponding elemental DMG,
        increase DMG.
        """
        if self.usage <= 0:
            # no usage, do nothing
            return value
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.damage_elemental_type != DamageElementalType.CRYO:
            # not corresponding elemental DMG, do nothing
            return value
        if value.position.area != ObjectPositionType.SKILL:
            # not character, do nothing
            return value
        # increase DMG
        assert mode == "REAL"
        value.damage += 1
        # decrease usage
        skill = match.get_object(value.position)
        if skill.skill_type == SkillType.NORMAL_ATTACK and self.talent_usage > 0:
            # decrease talent usage first
            self.talent_usage -= 1
        else:
            self.usage -= 1
        return value


class ChonghuasFrostField_3_3(ElementalInfusionTeamStatus, RoundTeamStatus):
    name: Literal["Chonghua's Frost Field"] = "Chonghua's Frost Field"
    desc: Literal["", "talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    talent_activated: bool = False

    infused_elemental_type: DamageElementalType = DamageElementalType.CRYO

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.talent_activated:
            self.desc = "talent"

    def renew(self, object: "ChonghuasFrostField_3_3"):
        super().renew(object)
        self.talent_activated = object.talent_activated
        if self.talent_activated:
            self.desc = "talent"

    def character_is_using_right_weapon(
        self, match: Any, position: ObjectPosition
    ) -> bool:
        """
        If position character using effected three weapon, return True.
        """
        character = match.player_tables[position.player_idx].characters[
            position.character_idx
        ]
        return character.weapon_type in [
            WeaponType.SWORD,
            WeaponType.CLAYMORE,
            WeaponType.POLEARM,
        ]

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        If weapon type not right, do not enhance.
        """
        if not self.character_is_using_right_weapon(match, value.position):
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When source weapon type is sword, claymore or polearm, and talent
        activated, +1 normal attack DMG.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not corresponding character, do nothing
            return value
        if not self.character_is_using_right_weapon(match, value.position):
            # not right weapon, do nothing
            return value
        # add damage
        assert mode == "REAL"
        value.damage += 1
        return value


class IceLotus_3_3(DefendTeamStatus):
    name: Literal["Ice Lotus"] = "Ice Lotus"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1


class FortunePreservingTalisman_4_0(UsageTeamStatus):
    name: Literal["Fortune-Preserving Talisman"] = "Fortune-Preserving Talisman"
    version: Literal["4.0"] = "4.0"
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        skill = match.get_object(event.action.position)
        if skill.name == "Adeptus Art: Preserver of Fortune":
            # is Preserver of Fortune, do nothing
            return []
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        character = match.player_tables[self.position.player_idx].characters[
            event.action.position.character_idx
        ]
        if character.damage_taken <= 0:
            # no damage, do nothing
            return []
        # heal character
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        ]


class CatClawShield_3_3(ShieldTeamStatus):
    name: Literal["Cat-Claw Shield"] = "Cat-Claw Shield"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1


class FlowingCicinShield_3_7(ShieldTeamStatus):
    name: Literal["Flowing Cicin Shield"] = "Flowing Cicin Shield"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1


register_class(
    Icicle_3_3
    | IcyQuill_3_7
    | ChonghuasFrostField_3_3
    | IceLotus_3_3
    | FortunePreservingTalisman_4_0
    | CatClawShield_3_3
    | FlowingCicinShield_3_7
)
